from ai2_kit.core.script import BashTemplate, BashStep, BashScript
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.log import get_logger
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import list_split, dict_nested_get, dump_json, dump_text
from ai2_kit.core.pydantic import BaseModel

from typing import List, Literal, Optional, Mapping, Sequence, Any
from pydantic import validator, root_validator
from dataclasses import dataclass
from string import Template
from allpairspy import AllPairs
from collections import defaultdict

import os
import itertools
import random
import ase.io


from .iface import BaseCllContext, ICllExploreOutput, TRAINING_MODE
from .constant import (
    LAMMPS_DUMP_DIR,
    LAMMPS_DUMP_SUFFIX,
    PRESET_LAMMPS_INPUT_TEMPLATE,
)
from .data import DataFormat, artifacts_to_ase_atoms
from .dpff import dump_dplr_lammps_data

logger = get_logger(__name__)


class CllLammpsInputConfig(BaseModel):
    n_wise: int = 0
    """
    The way of combining variants.
    if n_wise is less than 2 or greater than total fields, the full combination will be used.
    Or else, the n_wise combination will be used.
    It is strongly recommended to use n_wise when the full combination is too large.
    """

    explore_vars: Mapping[str, List[Any]]
    """
    Variants to be explore by full combination or n_wise combination.

    Variables defined here will become **LAMMPS variables**.
    If multiple value has been set for a variable,
    the cartesian product will be used to generate the combination.
    For example,

    ```yaml
    TEMP: [330, 430, 530]  # Can be a scalar, e.g. 330
    PRES: 1                # Can be a vector, e.g. [1, 2, 3]
    ```
    Then you can reference them in the LAMMPS input template as ${TEMP}, ${LAMBDA_f}, ${N_STEPS}, etc.
    """

    broadcast_vars: Mapping[str, Any] = dict()
    """
    Variants to be explore by broadcast.

    Variables defined here won't join the combination,
    but will be broadcasted to all combinations.

    This can be used to avoid combination explosion.

    ```yaml
    LAMBDA_f: [0.0, 0.25, 0.5, 0.75, 1.0]
    ```
    """
    preset_template: Optional[str] = None
    """
    Name of the preset template.
    """
    input_template: Optional[str] = None
    """
    LAMMPS input template file content.
    If set, the preset_template will be ignored.
    """
    template_vars: Mapping[str, Any] = dict()
    """
    input_template may provide extra injection points for user to inject custom settings.
    Those value could be set here.

    Those vars can be referenced in the LAMMPS input template as $$VAR_NAME.
    """

    plumed_config: Optional[str]
    """Plumed config file content."""

    system_files: List[str]
    """
    Artifacts key of lammps input data
    """

    ensemble: Optional[Literal['nvt', 'nvt-i', 'nvt-a', 'nvt-iso', 'nvt-aniso', 'npt', 'npt-t', 'npt-tri', 'nve', 'csvr']] = None
    fix_statement: Optional[str] = None

    no_pbc: bool = False
    nsteps: int
    timestep: float = 0.0005
    sample_freq: int = 100
    ignore_error: bool = False

    type_alias: Mapping[str, List[str]] = dict()
    '''
    Type alias for atoms. For example, if you want to distinguish ghost H and H of HF molecule from other H atoms,
    you can define the alias as follows:
    ```yaml
    type_alias:
        H: [ H_ghost, H_hf ]
    ```
    And then you can reference them in the LAMMPS input template, for example
    ```
    set atom 1 type ${H_hf}
    set atom 2 type ${H_ghost}
    ```
    '''

    @validator('explore_vars', pre=True)
    @classmethod
    def validate_explore_variants(cls, value):
        if not isinstance(value, dict):
            raise ValueError('explore_vars must be a dict')
        # override default values
        value = {
            'TAU_T': 0.1,
            'TAU_P': 0.5,
            'TIME_CONST': 0.1,
            **value,
        }
        result = {}
        for k, v in value.items():
            if not isinstance(v, list):
                v = [v]
            result[k] = v
        return result

    @root_validator()
    @classmethod
    def validate_domain(cls, values):
        ensemble = values.get('ensemble')
        if ensemble:
            no_pbc = values.get('no_pbc')
            if ensemble.startswith('npt') and no_pbc:
                raise ValueError('ensemble npt conflict with no_pcb')
            if not ensemble.startswith('npt'):
                logger.info('ensemble is not npt, force PRES to -1')
                values['explore_vars']['PRES'] = [-1]
        return values

    def assert_var(self, var: str, msg: str = ''):
        if not msg:
            msg = f'{var} is not defined in explore_vars or broadcast_vars'
        assert var in self.explore_vars or var in self.broadcast_vars, msg


class CllLammpsContextConfig(BaseModel):
    script_template: BashTemplate
    lammps_cmd: str = 'lmp'
    concurrency: int = 5


@dataclass
class CllLammpsInput:
    config: CllLammpsInputConfig
    type_map: List[str]
    mass_map: List[float]
    mode: TRAINING_MODE
    preset_template: str
    new_system_files: List[Artifact]
    dp_models: Mapping[str, List[Artifact]]
    dp_modifier: Optional[dict]
    dp_sel_type: Optional[List[int]]


@dataclass
class CllLammpsContext(BaseCllContext):
    config: CllLammpsContextConfig


@dataclass
class GenericLammpsOutput(ICllExploreOutput):
    model_devi_outputs: List[Artifact]

    def get_model_devi_dataset(self) -> List[Artifact]:
        return self.model_devi_outputs


async def cll_lammps(input: CllLammpsInput, ctx: CllLammpsContext):

    executor = ctx.resource_manager.default_executor
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)

    if len(input.new_system_files) > 0:
        data_files = input.new_system_files
    else:
        data_files = ctx.resource_manager.resolve_artifacts(input.config.system_files)

    assert len(data_files) > 0, 'no data files found'

    preset_template = input.config.preset_template
    if preset_template is None:
        if input.mode == 'fep-pka':
            preset_template = 'fep-pka'
        elif input.mode == 'fep-redox':
            preset_template = 'fep-redox'
        elif input.mode == 'dpff':
            preset_template = 'dpff'
        else:
            preset_template = 'default'

    tasks_dir, task_dirs = executor.run_python_fn(make_lammps_task_dirs)(
        combination_vars=input.config.explore_vars,
        broadcast_vars=input.config.broadcast_vars,
        data_files=[a.to_dict() for a in data_files],
        dp_models={k: [m.url for m in v] for k, v in input.dp_models.items()},
        n_steps=input.config.nsteps,
        timestep=input.config.timestep,
        sample_freq=input.config.sample_freq,
        no_pbc=input.config.no_pbc,
        n_wise=input.config.n_wise,
        ensemble=input.config.ensemble,
        fix_statement=input.config.fix_statement,
        preset_template=preset_template,
        input_template=input.config.input_template,
        plumed_config=input.config.plumed_config,
        extra_template_vars=input.config.template_vars,
        type_map=input.type_map,
        mass_map=input.mass_map,
        type_alias=input.config.type_alias,
        work_dir=work_dir,
        mode=input.mode,
        dp_modifier=input.dp_modifier,
        dp_sel_type=input.dp_sel_type,
        ai2_kit_cmd=f'{executor.python_cmd} -m ai2_kit.main',
    )

    # build scripts and submit
    base_cmd = f'{ctx.config.lammps_cmd} -i lammps.input'
    cmd = f'''if [ -f md.restart.* ]; then {base_cmd} -v restart 1; else {base_cmd} -v restart 0; fi'''

    # generate steps
    steps = []
    for task_dir in task_dirs:
        steps.append(BashStep(
            cwd=task_dir['url'], cmd=cmd, checkpoint='lammps', exit_on_error=not input.config.ignore_error))

    # submit jobs by the number of concurrency
    jobs = []
    for i, steps_group in enumerate(list_split(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=tasks_dir)
        jobs.append(job)

    await gather_jobs(jobs, max_tries=2)

    # build outputs
    outputs = []
    for task_dir in task_dirs:
        common = dict(url=task_dir['url'], executor=executor.name, format=DataFormat.LAMMPS_OUTPUT_DIR)
        if input.mode == 'fep-pka':
            # in fep-pka mode,
            # ini and fin states have different structures, so their lammps_dump_dir is different
            # their label method is different too, so we need to unpack `fep-ini` and `fep-fin` accordingly
            outputs += [
                Artifact.of(**common, attrs={
                    **task_dir['attrs'], 'model_devi_file': 'model_devi_ini.out', 'lammps_dump_dir': 'traj-ini',
                    **task_dir['attrs']['fep-ini'],
                }),
                Artifact.of(**common, attrs={
                    **task_dir['attrs'], 'model_devi_file': 'model_devi_fin.out', 'lammps_dump_dir': 'traj-fin',
                    **task_dir['attrs']['fep-fin'],
                    'ancestor': task_dir['attrs']['ancestor'] + '-fin',  # only fin needs
                }),
            ]
        elif input.mode == 'fep-redox':
            # in fep-redox mode,
            # ini and fin states have the same structure, so just use the default one,
            # but their label method is different, so we need to unpack `fep-ini` and `fep-fin` accordingly
            outputs += [
                Artifact.of(**common, attrs={
                    **task_dir['attrs'], 'model_devi_file': 'model_devi_ini.out',
                    **task_dir['attrs']['fep-ini'],
                }),
                Artifact.of(**common, attrs={
                    **task_dir['attrs'], 'model_devi_file': 'model_devi_fin.out',
                    **task_dir['attrs']['fep-fin'],
                    'ancestor': task_dir['attrs']['ancestor'] + '-fin',  # only fin needs
                }),
            ]
        else:
            outputs += [
                Artifact.of(**common, attrs={ **task_dir['attrs'] }),
            ]

    return GenericLammpsOutput(model_devi_outputs=outputs)



class LammpsInputTemplate(Template):
    delimiter = '$$'

def make_lammps_task_dirs(combination_vars: Mapping[str, Sequence[Any]],
                            broadcast_vars: Mapping[str, Sequence[Any]],
                            data_files: List[ArtifactDict],
                            dp_models: Mapping[str, List[str]],
                            n_steps: int,
                            timestep: float,
                            sample_freq: float,
                            no_pbc: bool,
                            n_wise: int,
                            ensemble: Optional[str],
                            fix_statement: Optional[str],
                            preset_template: str,
                            input_template: Optional[str],
                            plumed_config: Optional[str],
                            extra_template_vars: Mapping[str, Any],
                            type_map: List[str],
                            mass_map: List[float],
                            type_alias: Mapping[str, List[str]],
                            work_dir: str,
                            dp_modifier: Optional[dict],
                            dp_sel_type: Optional[List[int]],
                            mode: TRAINING_MODE,
                            ai2_kit_cmd: str,
                            ):
    # setup workspace
    input_data_dir = os.path.join(work_dir, 'input_data')
    tasks_dir = os.path.join(work_dir, 'tasks')
    for path in (input_data_dir, tasks_dir):
        os.makedirs(path, exist_ok=True)

    # create data files
    input_dataset = []
    atoms_list = artifacts_to_ase_atoms(data_files, type_map=type_map)
    for i, (artifact, atoms) in enumerate(atoms_list):
        #  create data file
        ancestor = artifact['attrs']['ancestor']
        data_file = os.path.join(input_data_dir, f'{ancestor}-{i:06d}.lammps.data')
        if mode == 'dpff':
            assert dp_modifier is not None and dp_sel_type is not None, 'dp_modifier & dp_sel_type is required for dpff mode'
            sys_charge_map = dp_modifier['sys_charge_map']
            model_charge_map = dp_modifier['model_charge_map']
            with open(data_file, 'w') as fp:
                dump_dplr_lammps_data(fp, atoms, type_map=type_map, sel_type=dp_sel_type,
                                        sys_charge_map=sys_charge_map, model_charge_map=model_charge_map)  # type: ignore
        else:
            ase.io.write(data_file, atoms, format='lammps-data', specorder=type_map)  # type: ignore
        input_dataset.append({
            'url': data_file,
            'attrs': artifact['attrs'],
        })

    # generate combinations of variants
    combination_fields: List[str] = ['DATA_FILE']
    combination_values: List[List[Any]] = [input_dataset]
    for k, v in combination_vars.items():
        combination_fields.append(k)
        combination_values.append(v)  # type: ignore
    if 1 < n_wise <= len(combination_fields):
        logger.info(f'using {n_wise}-wise combination')
        combinations = AllPairs(combination_values, n=n_wise)
    else:
        logger.info('using full combination')
        combinations = itertools.product(*combination_values)

    combinations = list(map(list, combinations))

    # broadcast broadcast_vars to all combinations
    for k in broadcast_vars.keys():
        # TODO: moving this check to pydatnic validator
        assert k not in combination_fields, f'broadcast_vars {k} is already in explore_vars'

    combination_fields.extend(broadcast_vars.keys())
    for i, combination in enumerate(combinations):
        for _vars in broadcast_vars.values():
            combination.append(_vars[i % len(_vars)])

    # generate tasks input
    task_dirs = []
    for i, combination in enumerate(combinations):
        lammps_vars = dict(zip(combination_fields, combination))
        template_vars = {
            **lammps_vars,
            'AI2KIT_CMD': ai2_kit_cmd,
        }

        # setup task dir
        task_dir = os.path.join(tasks_dir, f'{i:06d}')
        os.makedirs(os.path.join(task_dir, LAMMPS_DUMP_DIR), exist_ok=True)

        data_file = lammps_vars.pop('DATA_FILE')

        # override default values with data file attrs
        overridable_params: dict = dict_nested_get(data_file, ['attrs', 'lammps'], dict())  # type: ignore
        plumed_config = overridable_params.get('plumed_config', plumed_config)
        fix_statement = overridable_params.get('fix_statement', fix_statement)
        ensemble = overridable_params.get('ensemble', ensemble)
        type_alias = overridable_params.get('type_alias', type_alias)

        # be careful to override template_vars without changing the original dict
        extra_template_vars = {**extra_template_vars, **overridable_params.get('template_vars', dict())}

        # generate types related template vars
        types_template_vars = get_types_template_vars(type_map=type_map, mass_map=mass_map,
                                                        type_alias=type_alias, sel_type=dp_sel_type)

        ## build variables section
        lammps_vars['DATA_FILE'] = data_file['url']
        lammps_vars['N_STEPS'] = n_steps
        lammps_vars['THERMO_FREQ'] = sample_freq
        lammps_vars['DUMP_FREQ'] = sample_freq
        lammps_vars['SAMPLE_FREQ'] = sample_freq
        lammps_vars['DEFAULT_GROUP'] = 'all'

        if mode == 'dpff':
            assert dp_modifier is not None
            lammps_vars['DEFAULT_GROUP'] = 'real_atom'
            lammps_vars['EWALD_BETA'] = dp_modifier['ewald_beta']
        lammps_vars['DUMP_GROUP'] = lammps_vars['DEFAULT_GROUP']

        dump_json(lammps_vars, os.path.join(task_dir, 'debug.lammps_vars.json'))  # for debug
        template_vars['VARIABLES'] = _get_lammps_variables(lammps_vars)
        ## build init settings
        template_vars['INITIALIZE'] =  '\n'.join([
            'units           metal',
            'atom_style      %s' % ('full' if mode == 'dpff' else 'atomic'),
            'boundary ' + ('f f f' if no_pbc else 'p p p'),
        ])
        ## build read data section
        extra_types = sum(len(l) for l in type_alias.values())  # how many alias type are defined
        template_vars['READ_DATA'] = (
            '''if "${restart} > 0" '''
            '''then "read_restart md.restart.*" '''
            '''else "read_data ${DATA_FILE} extra/atom/types %s"''' % extra_types
        )

        ## build simulation section
        simulation = [
            '''if "${restart} == 0" then "velocity ${DEFAULT_GROUP} create ${TEMP} %d"''' % (random.randrange(10 ^ 6 - 1) + 1)
        ]

        if fix_statement is None:
            assert ensemble is not None, 'either fix_statement or ensemble is required'
            fix_statement = get_ensemble(ensemble, group='${DEFAULT_GROUP}')

        if mode == 'dpff':
            simulation.extend([
                'compute  real_temp real_atom temp',
                fix_statement,
                'fix_modify 1 temp real_temp',
                '',
            ])
        else:
            simulation.append(fix_statement)

        if plumed_config:
            plumed_config = LammpsInputTemplate(plumed_config).substitute(defaultdict(str), **template_vars)
            plumed_config_file = os.path.join(task_dir, 'plumed.input')
            dump_text(plumed_config, plumed_config_file)
            simulation.append(f'fix cll_plumed ${{DEFAULT_GROUP}} plumed plumedfile {plumed_config_file} outfile plumed.out')

        if no_pbc:
            simulation.extend([
                'velocity ${DEFAULT_GROUP} zero linear',
                'fix      fm ${DEFAULT_GROUP} momentum 1 linear 1 1 1',
            ])
        simulation.extend([
            'thermo_style custom step temp pe ke etotal press vol lx ly lz xy xz yz',
            'thermo       ${THERMO_FREQ}',
            'dump         1 ${DUMP_GROUP} custom ${DUMP_FREQ} %s/*%s id type x y z fx fy fz' % (LAMMPS_DUMP_DIR, LAMMPS_DUMP_SUFFIX),
            'restart      10000 md.restart',
        ])
        template_vars['SIMULATION'] = '\n'.join(simulation)
        ## build run section
        template_vars['RUN'] = '\n'.join([
            'timestep %f' % timestep,
            'run      ${N_STEPS} upto',
        ])

        dp_models_vars = _get_dp_models_variables(dp_models)
        template_vars = {**template_vars, **types_template_vars, **dp_models_vars, **extra_template_vars}
        dump_json(template_vars, os.path.join(task_dir, 'debug.template_vars.json'))

        if input_template is None:
            input_template = PRESET_LAMMPS_INPUT_TEMPLATE[preset_template]
        dump_text(input_template, os.path.join(task_dir, 'debug.input_template.txt'))
        lammps_input = LammpsInputTemplate(input_template).substitute(defaultdict(str),**template_vars)
        dump_text(lammps_input, os.path.join(task_dir, 'lammps.input'))

        # the `source` field is required as model_devi will use it to update init structures
        task_dirs.append({'url': task_dir,
                            'attrs': {
                                **data_file['attrs'],
                                'source': data_file['url'],
                                'efield': lammps_vars.get('EFIELD'),
                            }})  # type: ignore
    return tasks_dir, task_dirs


def get_types_template_vars(type_map: List[str], mass_map: List[float],
                            type_alias: Mapping[str, List[str]], sel_type: Optional[List[int]]):
    """
    generate template vars that related to type_map, mass_map, type_alias, sel_type

    the order of atom type index is:
    real atoms (defined in type_map), virtual atoms (defined in sel_type) and then alias (defined in type_alias)
    """
    template_vars = {}
    type_to_mass = dict(zip(type_map, mass_map))

    # new types gonna to be added
    ext_type_map = []
    ext_mass_map = []

    # handle sel_type (used by dplr)
    # sel_type must be handle before type_alias as they are defined in data file
    # while type_alias are defined in lammps script
    type_association = []
    if sel_type is not None:
        n_real_atom = len(type_map)
        for i, t in enumerate(sel_type):
            # add placeholder for sel_type
            ext_type_map.append(f'_X_{i}')
            ext_mass_map.append(1.0)
            # type association is to define relationship between virtual and real atom type in lammps
            type_association.extend([t + 1, n_real_atom + i + 1])
        template_vars['DPLR_TYPE_ASSOCIATION'] = ' '.join(map(str, type_association))

    # handle alias and fep special special rule
    alias_specorder = []
    alias_specorder_with_null = []
    for real_type, alias in type_alias.items():
        for t in alias:
            alias_specorder.append(real_type)
            if t.endswith('ghost') or t.endswith('null'):
                alias_specorder_with_null.append('NULL')
            else:
                alias_specorder_with_null.append(real_type)
            ext_type_map.append(t)
            ext_mass_map.append(type_to_mass[real_type])

    # inject group for dpff mode
    if sel_type is not None:
        real_atom_start = 1
        real_atom_end = real_atom_start + len(type_map)
        virtual_atom_end = real_atom_end + len(sel_type)
        alias_end = virtual_atom_end + len(ext_type_map) - len(sel_type)

        dpff_real_atom = [*range(real_atom_start, real_atom_end),  *range(virtual_atom_end, alias_end)]
        dpff_virtual_atom = range(real_atom_end, virtual_atom_end)

        template_vars['DPFF_REAL_ATOM'] = ' '.join(map(str, dpff_real_atom))
        template_vars['DPFF_VIRTUAL_ATOM'] = ' '.join(map(str, dpff_virtual_atom))
        template_vars['DPFF_GROUPS'] = '\n'.join([
            f'group real_atom    type {template_vars["DPFF_REAL_ATOM"]}',
            f'group virtual_atom type {template_vars["DPFF_VIRTUAL_ATOM"]}',
            f'neigh_modify    every 10 delay 0 check no exclude group real_atom virtual_atom',
        ])

    # SPECORDER is used to specify the order of types in the lammps data file
    # For example, if the complete type_map is [H, O, O_1, O_2, H_1, H_2],
    # then the specorder should be [H, O, O, O, H, H]
    specorder = type_map + alias_specorder
    fep_fin_specorder = type_map + alias_specorder_with_null

    # specorder in the format of H O H NULL, for lammps pair coeff input
    template_vars['SPECORDER'] = ' '.join(specorder)
    template_vars['SPECORDER_BASE'] = ' '.join(type_map)
    template_vars['FEP_INI_SPECORDER'] = template_vars['SPECORDER']
    template_vars['FEP_FIN_SPECORDER'] = ' '.join(fep_fin_specorder)

    # specorder in the format of ['H', 'O', 'H', 'NULL'], for ai2-kit command line input
    template_vars['SPECORDER_LIST'] = str(specorder)
    template_vars['SPECORDER_BASE_LIST'] = str(type_map)
    template_vars['FEP_INI_SPECORDER_LIST'] = template_vars['SPECORDER_LIST']
    template_vars['FEP_FIN_SPECORDER_LIST'] = str(fep_fin_specorder)

    # mass map is in the form of
    # variable   H               equal 1
    # variable   O               equal 2
    # variable   H_null          equal 3
    # mass ${H} 1.007
    # mass ${O} 15.999
    # mass ${H_null} 1.0
    template_vars['MASS_MAP_FULL'] = _get_masses(type_map + ext_type_map, mass_map + ext_mass_map)
    template_vars['MASS_MAP_BASE'] = _get_masses(type_map, mass_map)
    template_vars['MASS_MAP'] =  template_vars['MASS_MAP_FULL']
    return template_vars


def _get_type_map_vars(type_map: List[str]):
    return dict(zip(type_map, range(1, len(type_map) + 1)))


def _get_masses(type_map: List[str], mass_map: List[float]):
    lines = [
        _get_lammps_variables(_get_type_map_vars(type_map)),
        '',
    ]
    for t, m in zip(type_map, mass_map):
        lines.append(f'mass ${{{t}}} {m}')
    return '\n'.join(lines)


def _get_dp_models_variables(models: Mapping[str, List[str]]):
    vars = {}
    for k, v in models.items():
        prefix = 'DP_MODELS' if k == '' else f'DP_{k}_MODELS'
        vars[prefix] = ' '.join(v)
        for i, m in enumerate(v):
            vars[f'{prefix}_{i}'] = m
    return vars


def _get_lammps_variables(vars: Mapping[str, Any]):
    lines = []
    for k, v in vars.items():
        if isinstance(v, str):
            # TODO: should escape `v` in case of special characters
            line = f'variable    {k:16} string "{v}"'
        elif isinstance(v, list):  # vector or args
            line = f'variable    {k:16} string \"{" ".join(str(x) for x in v)}\"'
        else:
            line = f'variable    {k:16} equal {v}'
        lines.append(line)
    return '\n'.join(lines)


def get_ensemble(ensemble: str, group='all'):
    lines = []
    if ensemble in ('npt', 'npt-i', 'npt-iso',):
        lines.append('fix 1 %(group)s npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}')
    elif ensemble in ('npt-a', 'npt-aniso',):
        lines.append('fix 1 %(group)s npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P}')
    elif ensemble in ('npt-t', 'npt-tri',):
        lines.append('fix 1 %(group)s npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRES} ${PRES} ${TAU_P}')
    elif ensemble in ('nvt',):
        lines.append('fix 1 %(group)s nvt temp ${TEMP} ${TEMP} ${TAU_T}')
    elif ensemble in ('nve',):
        lines.append('fix 1 %(group)s nve')
    elif ensemble in ('csvr',):
        lines.append('fix 1 %(group)s nve')
        lines.append('fix 2 %(group)s temp/csvr ${TEMP} ${TEMP} ${TIME_CONST} %(seed)d')
    else:
        raise ValueError('unknown ensemble: ' + ensemble)
    return '\n'.join(lines) % {'group': group, 'seed': random.randrange(10^6 - 1) + 1}



import io
import re
import numpy as np


def cp2p_substitute_vars(string, vars: dict):
    # Define a regular expression for placeholders
    placeholder_pattern = re.compile(r"\$\{?(\w+)(?:-(\w+))?\}?")
    # Define a function to replace the placeholders with keyword arguments
    def replace_placeholder(match):
        # Get the variable name and the default value from the match object
        var_name = match.group(1)
        default_value = match.group(2)
        # Check if the variable name is in the keyword arguments
        if var_name in vars:
            # Return the value of the keyword argument
            return vars[var_name]
        else:
            # Check if there is a default value
            if default_value is not None:
                # Return the default value
                return default_value
            else:
                # Raise an exception if there is no default value
                raise ValueError(f"Missing keyword argument: {var_name}")
    # Return the string with the placeholders replaced
    return placeholder_pattern.sub(replace_placeholder, string)

def cp2k_process_macro(fp):
    # Define a regular expression for @SET directive with case insensitive flag
    set_pattern = re.compile(r"@SET\s+(\w+)\s+(.+)", re.IGNORECASE) # Added re.IGNORECASE flag
    # Initialize an empty dictionary for variables and a list for output lines
    variables = {}
    output_lines = []
    # Read line by line from the input file object
    for line in fp:
        # Strip whitespace and comments
        line = line.strip()
        line = line.split("#")[0]
        # Skip empty lines
        if not line:
            continue
        # Match @SET directive
        set_match = set_pattern.match(line)
        if set_match:
            # Get the variable name and value
            var_name = set_match.group(1)
            var_value = set_match.group(2).strip()

            # Check if the variable name is valid
            if var_name[0].isdigit():
                # Raise an exception if the variable name starts with a number
                raise ValueError(f"Invalid variable name: {var_name}")
            else:
                # Assign the value to the variable in the dictionary
                variables[var_name] = var_value
            # Skip the @SET line
            continue

        # Append the line to the output list
        output_lines.append(line)

    # Return the variables dictionary and the output list as a single string with newline characters
    return variables, "\n".join(output_lines)

def cp2k_parse_input(fp):
    # Initialize an empty dictionary and a stack
    output = {}
    stack = []
    # Open the input file and read line by line
    for line in fp:
        # Strip whitespace and comments
        line = line.strip()
        line = line.split("#")[0]
        # Skip empty lines
        if not line:
            continue
        # Check if the line starts with &
        if line.startswith("&"):
            # Get the keyword name and strip the trailing whitespace
            keyword = line[1:].strip()
            # Split the keyword by whitespace and check if the first token is END
            tokens = keyword.split()
            if tokens[0] and tokens[0].upper() == "END":
                # Pop the last section from the stack
                stack.pop()
            else:
                # If the stack is empty, add the keyword to the output dictionary
                if not stack:
                    output[keyword] = {}
                    stack.append(keyword)
                else:
                    # If the stack is not empty, get the current section from the output dictionary
                    current_section = output
                    for section in stack:
                        current_section = current_section[section]
                    # Add the keyword to the current section as a sub-dictionary
                    current_section[keyword] = {}
                    stack.append(keyword)
            # Continue to the next line
            continue
        # Split the line by whitespace
        tokens = line.split()
        # Get the value name and value
        value_name = tokens[0]
        value = " ".join(tokens[1:])
        # Get the current section from the output dictionary
        current_section = output
        for section in stack:
            current_section = current_section[section]
        # Add the value name and value to the current section
        current_section[value_name] = value
    return output


# TODO: handle coords
def cp2k_load_input(fp):
    variables, processed_text = cp2k_process_macro(fp)
    substituted_text = cp2p_substitute_vars(processed_text, variables)
    return cp2k_parse_input(io.StringIO(substituted_text))


def cp2k_loads_input(text):
    return cp2k_load_input(io.StringIO(text))


# TODO: handle coords
def cp2k_dumps_input(input_dict):
    # Initialize an empty list for output lines
    output_lines = []
    # Define a helper function to recursively dump the sections and values
    def dump_section(section_dict, indent=0):
        # Loop through the keys and values in the section dictionary
        for key, value in section_dict.items():
            # Check if the value is a sub-dictionary
            if isinstance(value, dict):
                # Add a line with the section name and indentation
                output_lines.append(" " * indent + f"&{key}")
                # Recursively dump the sub-section with increased indentation
                dump_section(value, indent + 3)
                # Add a line with the end of section and indentation
                output_lines.append(" " * indent + "&END")
            else:
                # Add a line with the value name and value and indentation
                output_lines.append(" " * indent + f"{key}  {value}")
    # Dump the input dictionary using the helper function
    dump_section(input_dict)
    # Return the output list as a single string with newline characters
    return "\n".join(output_lines)


def cp2k_dump_input(input_dict, fp):
    fp.write(cp2k_dumps_input(input_dict))


class LammpsData:
    """
    LammpsData is a class to read and write lammps data file.
    from: https://gitee.com/chiahsinchu/toolbox/blob/dev/toolbox/io/lammps.py
    """

    def __init__(self, atoms) -> None:
        self.atoms = atoms
        self._setup()

        self.angles = None
        self.bonds = None
        self.dihedrals = None

    def write(self, fp, **kwargs):
        specorder = kwargs.get("specorder", None)
        if specorder is not None:
            self.set_atype_from_specorder(specorder)
        atom_style = kwargs.get("atom_style", "full")

        header = self._make_header(fp.name)
        fp.write(header)
        body = self._make_atoms(atom_style)
        fp.write(body)
        if self.bonds is not None:
            fp.write("\nBonds\n\n")
            np.savetxt(fp, self.bonds, fmt="%d")
        if self.angles is not None:
            fp.write("\nAngles\n\n")
            np.savetxt(fp, self.angles, fmt="%d")
        if self.dihedrals is not None:
            fp.write("\nDihedrals\n\n")
            np.savetxt(fp, self.dihedrals, fmt="%d")

    def _make_header(self, out_file):
        cell = self.atoms.cell.cellpar()
        nat = len(self.atoms)
        s = "%s (written by toolbox by Jia-Xin Zhu)\n\n" % out_file
        s += "%d atoms\n" % nat
        s += "%d atom types\n" % len(np.unique(self.atoms.numbers))
        if self.bonds is not None:
            s += "%d bonds\n" % len(self.bonds)
            s += "%d bond types\n" % len(np.unique(self.bonds[:, 1]))
        if self.angles is not None:
            s += "%d angles\n" % len(self.angles)
            s += "%d angle types\n" % len(np.unique(self.angles[:, 1]))
        if self.dihedrals is not None:
            s += "%d dihedrals\n" % len(self.dihedrals)
            s += "%d dihedral types\n" % len(np.unique(self.dihedrals[:, 1]))
        s += "%.4f %.4f xlo xhi\n%.4f %.4f ylo yhi\n%.4f %.4f zlo zhi\n\n\n" % (
            0.0, cell[0], 0.0, cell[1], 0.0, cell[2])
        return s

    def _make_atoms(self, atom_style):
        """
        full atom_id res_id type q x y z
        atomic ...
        """
        return getattr(self, "_make_atoms_%s" % atom_style)()

    def _make_atoms_full(self):
        """
        full atom_id res_id type q x y z
        """
        s = "Atoms\n\n"
        for atom in self.atoms:
            ii = atom.index
            s += "%d %d %d %.6f %.6f %.6f %.6f\n" % (
                ii + 1, self.res_id[ii], self.atype[ii], self.charges[ii],
                self.positions[ii][0], self.positions[ii][1],
                self.positions[ii][2])
        return s

    def _make_atoms_atomic(self):
        pass

    def set_res_id(self, res_id):
        self.res_id = np.reshape(res_id, (-1))

    def set_atype(self, atype):
        self.atype = np.reshape(atype, (-1))

    def set_atype_from_specorder(self, specorder):
        atype = []
        for ii in self.atoms.get_chemical_symbols():
            atype.append(specorder.index(ii))
        self.atype = np.array(atype, dtype=np.int32) + 1

    def set_bonds(self, bonds):
        self.bonds = np.reshape(bonds, (-1, 4))

    def set_angles(self, angles):
        self.angles = np.reshape(angles, (-1, 5))

    def set_dihedral(self, dihedrals):
        self.dihedrals = np.reshape(dihedrals, (-1, 6))

    def _setup(self):
        self.positions = self.atoms.get_positions()
        if len(self.atoms.get_initial_charges()) > 0:
            self.charges = self.atoms.get_initial_charges().reshape(-1)
        else:
            self.charges = np.zeros(len(self.atoms))

        if hasattr(self, "res_id"):
            assert len(self.res_id) == len(self.atoms)
        else:
            self.res_id = np.zeros(len(self.atoms), dtype=np.int32)



from ai2_kit.core.executor import BaseExecutorConfig
from ai2_kit.core.artifact import ArtifactMap
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import load_yaml_files, merge_dict
from ai2_kit.core.resource_manager import ResourceManager
from ai2_kit.core.checkpoint import set_checkpoint_file, apply_checkpoint
from ai2_kit.core.pydantic import BaseModel
from ai2_kit.domain import (
    deepmd,
    iface,
    lammps,
    lasp,
    selector,
    cp2k,
    vasp,
    constant as const,
    updater,
)

from typing import Dict, List, Optional, Any
from dataclasses import dataclass
from fire import Fire

import asyncio
import itertools
import copy
import os

logger = get_logger(__name__)


class CllWorkflowExecutorConfig(BaseExecutorConfig):
    class Context(BaseModel):
        class Train(BaseModel):
            deepmd: deepmd.CllDeepmdContextConfig

        class Explore(BaseModel):
            lammps: Optional[lammps.CllLammpsContextConfig]
            lasp: Optional[lasp.CllLaspContextConfig]

        class Label(BaseModel):
            cp2k: Optional[cp2k.CllCp2kContextConfig]
            vasp: Optional[vasp.CllVaspContextConfig]

        train: Train
        explore: Explore
        label: Label

    context: Context


class WorkflowConfig(BaseModel):
    class General(BaseModel):
        type_map: List[str]
        mass_map: List[float]
        sel_type: Optional[List[str]]

        max_iters: int = 1
        mode: iface.TRAINING_MODE = 'default'
        update_explore_systems: bool = False

    class Label(BaseModel):
        cp2k: Optional[cp2k.CllCp2kInputConfig]
        vasp: Optional[vasp.CllVaspInputConfig]

    class Train(BaseModel):
        deepmd: deepmd.CllDeepmdInputConfig

    class Explore(BaseModel):
        lammps: Optional[lammps.CllLammpsInputConfig]
        lasp: Optional[lasp.CllLaspInputConfig]

    class Select(BaseModel):
        model_devi: selector.CllModelDeviSelectorInputConfig

    class Update(BaseModel):
        walkthrough: updater.CllWalkthroughUpdaterInputConfig

    general: General
    train: Train
    explore: Explore
    select: Select
    label: Label
    update: Update


class CllWorkflowConfig(BaseModel):

    executors: Dict[str, CllWorkflowExecutorConfig]
    artifacts: ArtifactMap
    workflow: Any  # Keep it raw here, it should be parsed later in iteration


def run_workflow(*config_files,
                 executor: Optional[str] = None,
                 path_prefix: Optional[str] = None,
                 checkpoint: Optional[str] = None):
    """
    Run Closed-Loop Learning (CLL) workflow to train Machine Learning Potential (MLP) models.

    Args:
        config_files: path of config files, should be yaml files, can be multiple, support glob pattern
        executor: name of executor, should be defined in config `executors` section
        path_prefix: path prefix for output
        checkpoint: checkpoint file
    """
    if checkpoint is not None:
        set_checkpoint_file(checkpoint)

    config_data = load_yaml_files(*config_files)
    config = CllWorkflowConfig.parse_obj(config_data)

    if executor not in config.executors:
        raise ValueError(f'executor {executor} is not found')
    if path_prefix is None:
        raise ValueError('path_prefix should not be empty')

    iface.init_artifacts(config.artifacts)
    resource_manager = ResourceManager(
        executor_configs=config.executors,
        artifacts=config.artifacts,
        default_executor=executor,
    )
    return asyncio.run(cll_mlp_training_workflow(config, resource_manager, executor, path_prefix))


async def cll_mlp_training_workflow(config: CllWorkflowConfig,
                                    resource_manager: ResourceManager,
                                    executor: str,
                                    path_prefix: str):
    context_config = config.executors[executor].context
    raw_workflow_config = copy.deepcopy(config.workflow)

    # output of each step
    label_output: Optional[iface.ICllLabelOutput] = None
    selector_output: Optional[iface.ICllSelectorOutput] = None
    train_output: Optional[iface.ICllTrainOutput] = None
    explore_output: Optional[iface.ICllExploreOutput] = None

    # cursor of update table
    update_cursor = 0

    # Start iteration
    for i in itertools.count(0):

        # parse workflow config
        workflow_config = WorkflowConfig.parse_obj(raw_workflow_config)
        shared_vars = precondition(workflow_config)

        if i >= workflow_config.general.max_iters:
            logger.info(f'Iteration {i} exceeds max_iters, stop iteration.')
            break

        # shortcut for type_map and mass_map
        type_map = workflow_config.general.type_map
        mass_map = workflow_config.general.mass_map

        # decide path prefix for each iteration
        iter_path_prefix = os.path.join(path_prefix, f'iters-{i:03d}')
        # prefix of checkpoint
        cp_prefix = f'iters-{i:03d}'

        # label
        if workflow_config.label.cp2k and context_config.label.cp2k:
            cp2k_input = cp2k.CllCp2kInput(
                config=workflow_config.label.cp2k,
                mode=workflow_config.general.mode,
                type_map=type_map,
                system_files=[] if selector_output is None else selector_output.get_model_devi_dataset(),
                initiated=i > 0,
            )
            cp2k_context = cp2k.CllCp2kContext(
                config=context_config.label.cp2k,
                path_prefix=os.path.join(iter_path_prefix, 'label-cp2k'),
                resource_manager=resource_manager,
            )
            label_output = await apply_checkpoint(f'{cp_prefix}/label-cp2k')(cp2k.cll_cp2k)(cp2k_input, cp2k_context)

        elif workflow_config.label.vasp and context_config.label.vasp:
            vasp_input = vasp.CllVaspInput(
                config=workflow_config.label.vasp,
                type_map=type_map,
                system_files=[] if selector_output is None else selector_output.get_model_devi_dataset(),
                initiated=i > 0,
            )
            vasp_context = vasp.CllVaspContext(
                config=context_config.label.vasp,
                path_prefix=os.path.join(iter_path_prefix, 'label-vasp'),
                resource_manager=resource_manager,
            )
            label_output = await apply_checkpoint(f'{cp_prefix}/label-vasp')(vasp.cll_vasp)(vasp_input, vasp_context)

        else:
            raise ValueError('No label method is specified')

        # return if no new data is generated
        if i > 0 and len(label_output.get_labeled_system_dataset()) == 0:
            logger.info("No new data is generated, stop iteration.")
            break

        # train
        if workflow_config.train.deepmd:
            deepmd_input = deepmd.CllDeepmdInput(
                config=workflow_config.train.deepmd,
                mode=workflow_config.general.mode,
                type_map=type_map,
                old_dataset=[] if train_output is None else train_output.get_training_dataset(),
                new_dataset=label_output.get_labeled_system_dataset(),
                sel_type=shared_vars.dp_sel_type,
                previous=[] if train_output is None else train_output.get_mlp_models(),
            )
            deepmd_context = deepmd.CllDeepmdContext(
                path_prefix=os.path.join(iter_path_prefix, 'train-deepmd'),
                config=context_config.train.deepmd,
                resource_manager=resource_manager,
            )
            train_output = await apply_checkpoint(f'{cp_prefix}/train-deepmd')(deepmd.cll_deepmd)(deepmd_input, deepmd_context)

        else:
            raise ValueError('No train method is specified')

        # explore
        new_explore_system_files = []
        if workflow_config.general.update_explore_systems and selector_output is not None:
            new_explore_system_files = selector_output.get_new_explore_systems()

        if workflow_config.explore.lammps and context_config.explore.lammps:
            lammps_input = lammps.CllLammpsInput(
                config=workflow_config.explore.lammps,
                mode=workflow_config.general.mode,
                type_map=type_map,
                mass_map=mass_map,
                dp_models={'': train_output.get_mlp_models()},
                preset_template='default',
                new_system_files=new_explore_system_files,
                dp_modifier=shared_vars.dp_modifier,
                dp_sel_type=shared_vars.dp_sel_type,
            )
            lammps_context = lammps.CllLammpsContext(
                path_prefix=os.path.join(iter_path_prefix, 'explore-lammps'),
                config=context_config.explore.lammps,
                resource_manager=resource_manager,
            )
            explore_output = await apply_checkpoint(f'{cp_prefix}/explore-lammps')(lammps.cll_lammps)(lammps_input, lammps_context)

        elif workflow_config.explore.lasp and context_config.explore.lasp:
            lasp_input = lasp.CllLaspInput(
                config=workflow_config.explore.lasp,
                type_map=type_map,
                mass_map=mass_map,
                models=train_output.get_mlp_models(),
                new_system_files=new_explore_system_files,
            )
            lasp_context = lasp.CllLaspContext(
                config=context_config.explore.lasp,
                path_prefix=os.path.join(iter_path_prefix, 'explore-lasp'),
                resource_manager=resource_manager,
            )
            explore_output = await apply_checkpoint(f'{cp_prefix}/explore-lasp')(lasp.cll_lasp)(lasp_input, lasp_context)

        else:
            raise ValueError('No explore method is specified')

        # select
        if workflow_config.select.model_devi:
            selector_input = selector.CllModelDeviSelectorInput(
                config=workflow_config.select.model_devi,
                model_devi_data=explore_output.get_model_devi_dataset(),
                model_devi_file=const.MODEL_DEVI_OUT,
                type_map=type_map,
            )
            selector_context = selector.CllModelDevSelectorContext(
                path_prefix=os.path.join(iter_path_prefix, 'selector-model-devi'),
                resource_manager=resource_manager,
            )
            selector_output = await apply_checkpoint(f'{cp_prefix}/selector-model-devi')(selector.cll_model_devi_selector)(selector_input, selector_context)
        else:
            raise ValueError('No select method is specified')

        # Update
        update_config = workflow_config.update.walkthrough

        # nothing to update because the table is empty
        if not update_config.table:
            continue
        # keep using the latest config when it reach the end of table
        if update_cursor >= len(update_config.table):
            continue

        # move cursor to next row if passing rate pass threshold
        if selector_output.get_passing_rate() > update_config.passing_rate_threshold:
            raw_workflow_config = merge_dict(copy.deepcopy(
                config.workflow), update_config.table[update_cursor])
            update_cursor += 1


@dataclass
class SharedVars():
    dp_modifier: Optional[dict] = None
    dp_sel_type: Optional[List[int]] = None


def precondition(workflow_cfg: WorkflowConfig) -> SharedVars:
    """
    precondition of workflow config to raise error early,
    and extra variables that may shared by multiple steps

    The known shared variables are:
      dp_modifier, which include vars sys_charge_map, model_charge_map, ewald_h, ewald_beta
      sel_type, which is suppose to be used in dplr/dpff mode
    """
    shared_vars = SharedVars()

    mode = workflow_cfg.general.mode
    type_map = workflow_cfg.general.type_map
    sel_type = workflow_cfg.general.sel_type

    if mode == 'dpff':
        assert sel_type is not None, 'sel_type should be specified in general config for dpff mode'
        shared_vars.dp_sel_type = [ type_map.index(t)  for t in sel_type ]

    deepmd_cfg = workflow_cfg.train.deepmd
    if deepmd_cfg is not None:
        if mode == 'dpff':
            modifier = deepmd_cfg.input_template['model'].get('modifier')
            assert modifier is not None, 'modifier should be specified in deepmd input template for dpff mode'
            shared_vars.dp_modifier = modifier
        elif mode == 'fep-redox':
            assert deepmd_cfg.input_template['model']['fitting_net']['numb_fparam'] == 1, 'numb_fparam should be 1 for fep-redox/fep-pka mode'

    lammps_cfg = workflow_cfg.explore.lammps
    if lammps_cfg is not None:
        if mode == 'dpff':
            lammps_cfg.assert_var('EFIELD')
            lammps_cfg.assert_var('KMESH')
            efield = lammps_cfg.explore_vars.get('EFIELD', lammps_cfg.broadcast_vars.get('EFIELD'))
            assert all([isinstance(item, list) for item in efield ]), 'EFIELD should be a list of vector'  # type: ignore
        elif mode in ['fep-redox', 'fep-pka']:
            lammps_cfg.assert_var('LAMBDA_f')

    return shared_vars


if __name__ == '__main__':
    # use python-fire to parse command line arguments
    Fire(run_workflow)



from ase import Atoms
import ase.io

from typing import List, Optional
from collections import OrderedDict
import shlex
import os

from ai2_kit.tool.dpdata import register_data_types
from ai2_kit.core.util import list_split, expand_globs, cmd_with_checkpoint
from ai2_kit.domain.cp2k import dump_coord_n_cell
from ai2_kit.core.log import get_logger


register_data_types()
logger = get_logger(__name__)


def ensure_file_exists(file_path: str):
    assert os.path.isfile(file_path), f'{file_path} is not a file'
    return file_path


class Cp2kLabelTaskBuilder:
    """
    This a builder class to generate cp2k labeling task.
    """
    def __init__(self):
        self._atoms_list: List[Atoms] = []
        self._task_dirs: List[str] = []
        self._cp2k_inputs: OrderedDict = OrderedDict()

    def add_system(self, *file_path_or_glob: str, **kwargs):
        """
        Add system files to label

        :param file_path_or_glob: system files or glob pattern
        :param kwargs: kwargs for ase.io.read
        """
        files = expand_globs(file_path_or_glob)
        if len(files) == 0:
            raise FileNotFoundError(f'No file found for {file_path_or_glob}')
        for file in files:
            self._ase_read(file, **kwargs)
        return self

    def add_cp2k_input(self, file_path: str, tag: str):
        """
        Add cp2k input file for dipole or polarizability calculation
        """
        assert tag not in self._cp2k_inputs, f'Tag {tag} already exists'
        self._cp2k_inputs[tag] = ensure_file_exists(file_path)
        return self

    def make_task(self, out_dir: str):
        """
        Make task dirs for cp2k labeling (prepare systems and cp2k input files)
        :param out_dir: output dir
        """
        assert self._atoms_list, 'No system files added'
        assert not self._task_dirs, 'Task dirs already generated'

        task_dirs = []
        for i, atoms in enumerate(self._atoms_list):
            task_name = f'{i:06d}'
            task_dir = os.path.join(out_dir, task_name)
            os.makedirs(task_dir, exist_ok=True)
            # dump coord_n_cell.inc, so that we can use @include to read system
            cp2k_sys_file = os.path.join(task_dir, 'coord_n_cell.inc')
            with open(cp2k_sys_file, 'w') as f:
                dump_coord_n_cell(f, atoms)
            # dump system.xyz for cp2k,
            # so that use can use COORD_FILE_NAME to read system
            xyz_sys_file = os.path.join(task_dir, 'system.xyz')
            ase.io.write(xyz_sys_file, atoms, format='extxyz', append=True)
            # link cp2k input files to tasks dirs
            for tag, file_path in self._cp2k_inputs.items():
                link_path = os.path.join(task_dir, tag + '.inp')
                os.system(f'ln -sf {os.path.abspath(file_path)} {link_path}')
            task_dirs.append(task_dir)
        self._task_dirs = task_dirs
        logger.info(f'Generated {len(task_dirs)} task dirs')
        return self

    def make_batch(self,
                   prefix: str = 'cp2k-batch-{i:02d}.sub',
                   concurrency: int = 5,
                   template: Optional[str] = None,
                   ignore_error: bool = False,
                   cp2k_cmd: str = 'cp2k.psmp'):
        """
        Make batch script for cp2k labeling

        :param prefix: prefix of batch script
        :param concurrency: concurrency of tasks
        :param template: template of batch script
        :param ignore_error: ignore error
        :param cmd: command to run cp2k
        """
        assert self._atoms_list, 'No system files added, please call add_system first'
        assert self._task_dirs, 'No task dirs generated, please call make_dirs first'
        if template is None:
            template = '#!/bin/bash'
        else:
            with open(template, 'r') as f:
                template = f.read()
        template += '\n[ -n "$PBS_O_WORKDIR" ] && cd $PBS_O_WORKDIR]\n'

        os.makedirs(os.path.dirname(prefix), exist_ok=True)
        for i, task_group in enumerate(list_split(self._task_dirs, concurrency)):
            if not task_group:
                continue
            batch = [template]
            batch.append("for _WORK_DIR_ in \\")
            # use shell for loop to run tasks
            for task_dir in task_group:
                batch.append(f'  {shlex.quote(task_dir)} \\')
            batch.extend([
                f'; do',
                f'pushd $_WORK_DIR_ || exit 1',
                '#' * 80,
                *[
                    cmd_with_checkpoint(f'{cp2k_cmd} -i {tag}.inp &> {tag}.out', f'{tag}.done', ignore_error)
                    for tag in self._cp2k_inputs.keys()
                ],
                '#' * 80,
                f'popd',
                f'done'
            ])
            batch_file = prefix.format(i=i)
            with open(batch_file, 'w', encoding='utf-8') as f:
                f.write('\n'.join(batch))
            logger.info(f'Write batch script to {batch_file}')
        return self

    def done(self):
        """
        Mark the end of commands chain
        """

    def _ase_read(self, filename: str, **kwargs):
        kwargs.setdefault('index', ':')
        data = ase.io.read(filename, **kwargs)
        if not isinstance(data, list):
            data = [data]
        self._atoms_list += data
        logger.info(f'Loaded {len(data)} systems from {filename}, total {len(self._atoms_list)} systems')


cmd_entry = {
        'cp2k-labeling': Cp2kLabelTaskBuilder,
    }


if __name__ == '__main__':
    from fire import Fire
    Fire(cmd_entry)



import dpdata
from dpdata.unit import econvs
from ase import Atoms
import re
import numpy as np
from typing import List

from .util import LammpsData

def __expore_remote():

    def set_dplr_ext_from_cp2k_output(dp_sys: dpdata.LabeledSystem,
                                      cp2k_output: str,
                                      wannier_file: str,
                                      type_map: List[str],
                                      sys_charge_map: List[int],
                                      model_charge_map: List[int],
                                      ewald_h: float,
                                      ewald_beta: float,
                                      ext_efield,
                                      sel_type: List[int],
                                      ):

        import ase.io
        wannier_atoms = ase.io.read(wannier_file)
        with open(cp2k_output, 'r') as fp:
            raw_energy = get_cp2k_output_total_energy(fp)

        ext_efield = np.reshape(ext_efield, [1, 3])
        natoms = dp_sys.get_natoms()
        nframes = dp_sys.get_nframes()
        assert nframes == 1, 'Only support one frame'

        symbols = np.array(dp_sys.data["atom_names"])[dp_sys.data["atom_types"]]
        sel_ids = get_sel_ids(dp_sys, type_map, sel_type)

        atomic_dipole, extended_coords = get_atomic_dipole(dp_sys, sel_ids, wannier_atoms)
        # get extended charges
        extended_charges = np.zeros(len(extended_coords))
        for symbol, ion_charge in zip(type_map, sys_charge_map):
            extended_charges[np.where(symbols == symbol)[0]] = ion_charge
        sel_symbols = symbols[sel_ids]
        for ii, wannier_charge in zip(sel_type, model_charge_map):
            sel_symbol = type_map[ii]
            extended_charges[np.where(sel_symbols == sel_symbol)[0] + natoms] = wannier_charge
        total_dipole = np.sum(extended_coords * extended_charges.reshape(-1, 1), axis=0)
        energy = raw_energy - np.inner(total_dipole, ext_efield.reshape(3))
        dp_sys.data['energies'] = np.reshape(energy, [-1])
        pbc_efield = get_pbc_atomic_efield(dp_sys, extended_coords, extended_charges, ewald_h, ewald_beta)
        # natoms * 3
        _efield = pbc_efield + ext_efield
        aparam = np.linalg.norm(_efield, axis=-1)
        efield = np.array(_efield) / aparam.reshape(natoms, 1)

        dp_sys.data['ext_efield'] = ext_efield.reshape(nframes, 3)
        dp_sys.data['efield'] = efield.reshape(nframes, natoms, 3)
        dp_sys.data['aparam'] = aparam.reshape(nframes, natoms, -1)
        dp_sys.data['atomic_dipole'] = atomic_dipole.reshape(nframes, -1)
        return dp_sys

    def get_cp2k_output_total_energy(fp):
        start_pattern = r"  Total charge density g-space grids:"
        end_pattern = r"  Total energy:"
        start_pattern = re.compile(start_pattern)
        end_pattern = re.compile(end_pattern)

        flag = False
        data_lines = []
        nloop = 0
        for line in fp:
            line = line.strip('\n')
            if start_pattern.match(line):
                flag = True
            elif flag and end_pattern.match(line):
                flag = False
                nloop += 1

            if flag is True:
                data_lines.append(line)
        data_lines = np.reshape(data_lines, (nloop, -1))

        tot_e = 0.
        # grep the data from the last iteration
        for kw in data_lines[-1, 2:-1]:
            kw = kw.split(":")
            k = kw[0].strip(' ')
            v = float(kw[1]) * econvs["hartree"]
            if k != "Fermi energy":
                tot_e += v
        return tot_e


    def get_sel_ids(dp_sys, type_map, sel_type):
        symbols = np.array(dp_sys.data["atom_names"])[dp_sys.data["atom_types"]]
        sel_ids = [np.where(symbols == type_map[ii])[0] for ii in sel_type]
        return np.concatenate(sel_ids)


    def get_atomic_dipole(dp_sys, sel_ids, wannier_atoms, wannier_cutoff = 1.0):
        from ase.geometry.cell import cell_to_cellpar
        from MDAnalysis.lib.distances import distance_array, minimize_vectors

        coords = dp_sys.data['coords'].reshape(-1, 3)
        cellpar = cell_to_cellpar(dp_sys.data['cells'].reshape(3, 3))
        extended_coords = coords.copy()

        ref_coords = coords[sel_ids].reshape(-1, 3)
        e_coords = wannier_atoms.get_positions()
        dist_mat = distance_array(ref_coords, e_coords, box=cellpar)
        atomic_dipole = []
        for ii, dist_vec in enumerate(dist_mat):
            mask = (dist_vec < wannier_cutoff)
            cn = np.sum(mask)
            assert cn == 4
            wc_coord_rel = e_coords[mask] - ref_coords[ii]
            wc_coord_rel = minimize_vectors(wc_coord_rel, box=cellpar)
            _atomic_dipole = wc_coord_rel.mean(axis=0)
            atomic_dipole.append(_atomic_dipole)
            wc_coord = _atomic_dipole + ref_coords[ii]
            extended_coords = np.concatenate((extended_coords, wc_coord.reshape(1, 3)), axis=0)
        atomic_dipole = np.reshape(atomic_dipole, (-1, 3))
        assert atomic_dipole.shape[0] == len(sel_ids)
        return atomic_dipole, extended_coords


    def get_pbc_atomic_efield(dp_sys,
                                extended_coords,
                                extended_charges,
                                ewald_h,
                                ewald_beta,
                                ):
        from deepmd.infer.ewald_recp import EwaldRecp

        er = EwaldRecp(ewald_h, ewald_beta)
        natoms = dp_sys.get_natoms()
        cell = dp_sys.data['cells']
        e, f, v = er.eval(extended_coords.reshape(1, -1),
                            extended_charges.reshape(1, -1),
                            cell.reshape(1, -1))
        extended_efield = f.reshape(-1, 3) / extended_charges.reshape(-1, 1)
        # natoms * 3
        pbc_efield = extended_efield[:natoms]
        return pbc_efield

    def get_sel_type(model_path) -> List[int]:
        try:
            from deepmd.infer import DeepDipole
        except ImportError:
            raise ImportError('deepmd-kit is not installed')

        dp = DeepDipole(model_path)
        return [t for t in dp.tselt]


    def build_sel_type_assertion(sel_type, model_path: str, py_cmd='python'):
        return f'''{py_cmd} -c "from deepmd.infer import DeepDipole;dp = DeepDipole({repr(model_path)});assert{repr(sel_type)}==[t for t in dp.tselt]"'''


    def dump_dplr_lammps_data(fp, atoms: Atoms, type_map: List[str], sel_type: List[int],
                              sys_charge_map: List[float], model_charge_map: List[float]):

        assert len(type_map) == len(sys_charge_map), f'type_map {type_map} and sys_charge_map {sys_charge_map} must have the same length'
        assert len(sel_type) == len(model_charge_map), f'sel_type {sel_type} and model_charge_map {model_charge_map} must have the same length'

        new_atoms = atoms.copy()
        vtypes = get_unused_symbols(type_map, len(sel_type))

        r_atom_ids = []
        v_atom_ids = []

        # add virtual atoms
        for atype, vtype in zip(sel_type, vtypes):
            # create virtual atoms from the original atom type
            v_atoms = atoms[atoms.symbols == type_map[atype]]
            v_atoms.set_chemical_symbols([vtype] * len(v_atoms))  # type: ignore

            # create bonds between atom and its virtual atoms
            r_atom_ids.extend(np.where(atoms.symbols == type_map[atype])[0])
            v_atom_ids.extend(np.arange(len(v_atoms)) + len(new_atoms))  # type: ignore
            # this should be last
            new_atoms += v_atoms

        # build charges
        charges = np.zeros(len(new_atoms))
        for atype, charge in zip(type_map + vtypes, sys_charge_map + model_charge_map):
            charges[new_atoms.symbols==atype] = charge
        # build bonds
        n_bonds = len(r_atom_ids)
        # note: the index of lammps start from 1
        bonds = np.array([
            np.arange(n_bonds) + 1,  # bond id
            [1] * n_bonds,  # bond type
            np.array(r_atom_ids) + 1,  # bond left
            np.array(v_atom_ids) + 1,  # bond right
        ], dtype=int).T

        # write lammps data
        lmp_data = LammpsData(new_atoms)
        lmp_data.charges = charges
        lmp_data.bonds = bonds
        lmp_data.write(fp, atom_style='full', specorder=type_map + vtypes)

    def get_unused_symbols(used_symbols: List[str], size: int):
        """
        Get unused symbols from ase

        :param used_symbols: symbols that are already used
        :param size: number of unused symbols to get
        """
        from ase.data import chemical_symbols

        unused_symbols = []
        for symbol in reversed(chemical_symbols):
            if symbol not in used_symbols:
                unused_symbols.append(symbol)
            if len(unused_symbols) == size:
                break
        return unused_symbols

    return (
        get_sel_type,
        set_dplr_ext_from_cp2k_output,
        dump_dplr_lammps_data,
        build_sel_type_assertion,
    )

(
    get_sel_type,
    set_dplr_ext_from_cp2k_output,
    dump_dplr_lammps_data,
    build_sel_type_assertion,
) = __expore_remote()



from ai2_kit.core.artifact import ArtifactDict

from typing import List, Tuple, Optional
from ase import Atoms

import ase.io
import os


class DataFormat:
    # customize data format
    CP2K_OUTPUT_DIR = 'cp2k/output_dir'
    VASP_OUTPUT_DIR = 'vasp/output_dir'
    LAMMPS_OUTPUT_DIR = 'lammps/output_dir'
    DEEPMD_OUTPUT_DIR = 'deepmd/output_dir'
    DEEPMD_MODEL = 'deepmd/model'
    DEEPMD_NPY = 'deepmd/npy'
    LASP_LAMMPS_OUT_DIR ='lasp+lammps/output_dir'

    # data format of dpdata
    CP2K_OUTPUT = 'cp2k/output'
    VASP_XML = 'vasp/xml'

    # data format of ase
    EXTXYZ = 'extxyz'
    VASP_POSCAR = 'vasp/poscar'


def get_data_format(artifact: dict) -> Optional[str]:
    """
    Get (or guess) data type from artifact dict
    Note: The reason of using dict instead of Artifact is Artifact is not pickleable
    """
    url = artifact.get('url')
    assert isinstance(url, str), f'url must be str, got {type(url)}'

    file_name = os.path.basename(url)
    format = artifact.get('format')
    if format and isinstance(format, str):
        return format  # TODO: validate format
    if file_name.endswith('.xyz'):
        return DataFormat.EXTXYZ
    if 'POSCAR' in file_name:
        return DataFormat.VASP_POSCAR
    return None


def artifacts_to_ase_atoms(artifacts: List[ArtifactDict], type_map: List[str]) -> List[Tuple[ArtifactDict, Atoms]]:
    results = []
    for a in artifacts:
        data_format = get_data_format(a)  # type: ignore
        url = a['url']
        if data_format == DataFormat.VASP_POSCAR:
            atoms_list = ase.io.read(url, ':', format='vasp')
        elif data_format == DataFormat.EXTXYZ:
            atoms_list = ase.io.read(url, ':', format='extxyz')
        else:
            raise ValueError(f'unsupported data format: {data_format}')
        results.extend((a, atoms) for atoms in atoms_list)
    return results


def ase_atoms_to_cp2k_input_data(atoms: Atoms) -> Tuple[List[str], List[List[float]]]:
    coords = [atom.symbol + ' ' + ' '.join(str(x) for x in atom.position) for atom in atoms] # type: ignore
    cell = [list(row) for row in atoms.cell]  # type: ignore
    return (coords, cell)


def convert_to_lammps_input_data(systems: List[ArtifactDict], base_dir: str, type_map: List[str]):
    data_files = []
    atoms_list = artifacts_to_ase_atoms(systems, type_map=type_map)
    for i, (artifact, atoms) in enumerate(atoms_list):
        data_file = os.path.join(base_dir, f'{i:06d}.lammps.data')
        ase.io.write(data_file, atoms, format='lammps-data', specorder=type_map)  # type: ignore
        data_files.append({
            'url': data_file,
            'attrs': artifact['attrs'],
        })
    return data_files



import os
import sys
import shutil
import shlex
from typing import Optional

from ai2_kit.core.util import ensure_dir, expand_globs, list_split
from ai2_kit.core.log import get_logger

logger = get_logger(__name__)


class BatchHelper:
    """
    A toolkit to help generate batch scripts.
    """

    def run_cmd(self, *work_dirs: str, cmd: str):
        """
        Run command in each work directory.

        :param work_dirs: path or glob of work directories
        :param cmd: command to run, use {work_dir} to represent the work directory
        """

        paths = expand_globs(work_dirs)

        for path in paths:
            assert os.path.isdir(path), f'{path} is not a directory'
            _cmd = f"cd {shlex.quote(path)} && {cmd.format(work_dir=path)}"
            os.system(_cmd)
            logger.info(f'Run command: {_cmd}')
        return self

    def map_path(self, *sources: str, target: str, copy = False):
        """
        Map source files or directory to target path, use link by default.

        :param sources: path or glob of source files or directories
        :param target: target path, use {i} to represent the index of the source path,
        or {basename} to represent the basename of the source path
        :param copy: use copy instead of link
        """

        paths = expand_globs(sources)

        for i, path in enumerate(paths):
            target_path = target.format(i=i, basename=os.path.basename(path))
            ensure_dir(target_path)
            if copy:
                if os.path.isdir(path):
                    shutil.copytree(path, target_path)
                    logger.info(f'Copy directory {path} to {target_path}')
                else:
                    shutil.copy(path, target_path)
                    logger.info(f'Copy file {path} to {target_path}')
            else:
                path = os.path.abspath(path)
                os.symlink(path, target_path)
                logger.info(f'Link {path} to {target_path}')
        return self

    def gen_batches(self, *work_dirs: str,
                    out: str,
                    cmd: Optional[str]=None,
                    concurrency: int = 1,
                    header_file: Optional[str]=None,
                    suppress_error: bool = False,
                    checkpoint: bool = True,
                    checkpoint_file: str = 'done.ckpt',
                    rel_path: bool = False,
                    ):
        """
        Generate batch scripts for each work directory.

        This command will apply `cmd` to each work directory and generate batch scripts according to `concurrency`.

        :param work_dirs: path or glob of work directories
        :param out: path to write batch scripts, use {i} to represent the index of concurrent job
        :param cmd: command to run, if None, will read from stdin, use {word_dir} to represent the word directory,
        use {i} to represent the index of concurrent job
        :param concurrency: number of concurrent jobs, decide the number of batch scripts to generate, if 0, will generate one batch script for each work directory
        :param header_file: path to header file, will be added to the beginning of each batch script
        :param suppress_error: if True, will add `set -e` to the beginning of each batch script
        :param checkpoint: if True, will add checkpoint to each batch script, and skip the work directory if checkpoint exists
        :param checkpoint_file: checkpoint file name
        :param rel_path: if True, will use relative path in batch script
        """
        _work_dirs = expand_globs(work_dirs)
        # read cmd
        if cmd is None:
            cmd = sys.stdin.read()

        # read template
        header = '#!/bin/bash'
        if header_file is not None:
            with open(header_file, encoding='utf-8') as f:
                header = f.read()

        # generate batch scripts
        if concurrency <= 0:
            concurrency = len(_work_dirs)

        for i, job_group in enumerate(list_split(_work_dirs, concurrency)):
            batch = [ header ]
            if not suppress_error:
                batch.append('set -e')

            batch.append("for work_dir in \\")
            # generate batch script
            for work_dir in job_group:
                assert os.path.isdir(work_dir), f'{work_dir} is not a directory'
                if not rel_path:
                    work_dir = os.path.abspath(work_dir)
                batch.append(f'  {shlex.quote(work_dir)} \\')
            batch.extend([
                f'  ; do',
                f'    pushd $work_dir',
            ])
            if checkpoint:
                batch.extend([
                    f'    if [ -f {checkpoint_file}]; then',
                    f'      echo "hit checkpoint, skip"',
                    f'      continue',
                    f'    fi',
                ])
            batch.extend([
                f'    {cmd}',
                f'    touch {checkpoint_file}',
                f'    popd',
                f'  done'
            ])

            # write batch script
            out_path = out.format(i=i)
            ensure_dir(out_path)
            with open(out_path, 'w', encoding='utf-8') as f:
                f.write('\n'.join(batch))
            logger.info(f'Write batch script to {out_path}')
        return self

    def __str__(self) -> str:
        # suppress fire help message
        return ''




# Suppress numba warning of umap.
# Ref: https://github.com/lmcinnes/umap/issues/252
from numba.core.errors import NumbaDeprecationWarning
import warnings
warnings.filterwarnings("ignore", category=NumbaDeprecationWarning)

from asaplib.data.xyz import ASAPXYZ
from asaplib.cli.func_asap import cluster_process, set_reducer
from asaplib.hypers.hyper_soap import universal_soap_hyper
from asaplib.hypers.hyper_acsf import universal_acsf_hyper
from asaplib.reducedim.dim_reducer import Dimension_Reducers
from asaplib.cluster.ml_cluster_fit import LAIO_DB, sklearn_DB

from scipy.spatial.distance import cdist

import numpy as np



def get_descriptor(asapxyz: ASAPXYZ, setting: dict, keep_atomic=False, n_process=4, path_prefix='./asap-descriptor'):

    descriptor_spec = get_descriptor_spec(asapxyz, setting)
    asapxyz.compute_global_descriptors(
        desc_spec_dict = descriptor_spec,
        keep_atomic = keep_atomic,
        n_process = n_process,
    )
    asapxyz.save_state(path_prefix)

    global_descriptors = asapxyz.fetch_computed_descriptors(list(descriptor_spec.keys()))
    atomic_descriptors = asapxyz.fetch_computed_atomic_descriptors(list(descriptor_spec.keys())) if keep_atomic else None
    return global_descriptors, atomic_descriptors


def get_descriptor_spec(asapxyz: ASAPXYZ, setting: dict):
    descriptor_type, params = None, None
    for key in ['soap', 'acsf', 'cm']:
        params = setting.get(key, None)
        if params is not None:
            descriptor_type = key
            break
    else:
        raise ValueError('Unknown descriptor type')


    if descriptor_type == 'cm':
        return {'cm': {'type': 'CM'}}

    reducer_spec = dict(set_reducer(
        params['reducer_type'],
        params['element_wise'],
        params['zeta'],
    ))

    if descriptor_type == 'acsf':
        atomic_descriptor_spec = get_acsf_atomic_descriptor_spec(asapxyz, params)
    elif descriptor_type == 'soap':
        atomic_descriptor_spec = get_soap_atomic_descriptor_spec(asapxyz, params)

    descriptor_spec = {}
    for k, v in atomic_descriptor_spec.items():  # type: ignore
        descriptor_spec[k] = {
            'atomic_descriptor': {k: v},
            'reducer_function': reducer_spec,
        }
    return descriptor_spec


def get_soap_atomic_descriptor_spec(asapxyz: ASAPXYZ, params: dict):
    preset = params['preset']
    if preset is None:
        atomic_descriptor_spec = {
            'soap1': {
                'type': 'SOAP',
                'cutoff': params['r_cut'],
                'n': params['n_max'],
                'l': params['l_max'],
                'atom_gaussian_width': params['sigma'],
            }
        }
    else:
        atomic_descriptor_spec = universal_soap_hyper(asapxyz.get_global_species(), preset, dump=False)

    for k in atomic_descriptor_spec.keys():
        atomic_descriptor_spec[k]['rbf'] = params['rbf']
        atomic_descriptor_spec[k]['crossover'] = params['crossover']

    return atomic_descriptor_spec


def get_acsf_atomic_descriptor_spec(asapxyz: ASAPXYZ, params: dict):
    preset = params['preset']
    facsf_param = params['r_cut'] if preset is None else preset
    return universal_acsf_hyper(asapxyz.get_global_species(), facsf_param, dump=False)


def reduce_dimension(descriptors: np.ndarray, setting: dict):
    reducer = Dimension_Reducers(setting)
    reduced_descriptors = reducer.fit_transform(descriptors)
    return reduced_descriptors


def get_dbscan_trainer(descriptors: np.ndarray, metric='euclidean', eps=None, min_samples=2, eval_sample=50):
    if eps is None:
        n = len(descriptors)
        samples = descriptors[np.random.choice(n, min(n, eval_sample), replace=False)]
        # FIXME: the method to estimate eps is strange
        # FIXME: this will be broken when len of descriptors small
        eps = np.percentile(cdist(samples, descriptors, metric), min(100 * 10. / n, 99))  # type: ignore
    return sklearn_DB(eps, min_samples, metrictype=metric)


def get_laio_db_trainer(**kwargs):
    return LAIO_DB(**kwargs)


def get_trainer(descriptor: np.ndarray, setting: dict):
    dbscan_setting = setting.get('dbscan', None)
    if dbscan_setting is not None:
        return get_dbscan_trainer(descriptor, **dbscan_setting)
    laiodb_setting = setting.get('laiodb', None)
    if laiodb_setting is not None:
        return get_laio_db_trainer(**laiodb_setting)
    raise ValueError('Unknown trainer type')


def get_cluster(asapxyz: ASAPXYZ, descriptors: np.ndarray, trainer, path_prefix='./asap-cluster'):
    options = {
        'prefix': path_prefix,
        'savexyz': False,
        'savetxt': False,
    }
    labels: np.ndarray = cluster_process(asapxyz, trainer, descriptors, options)  # type: ignore

    groups = {}
    index = np.arange(len(labels))
    for label in set(labels):
        groups[label] = index[labels == label]
    return groups



from ai2_kit.core.util import merge_dict, wait_for_change, load_json, dump_json
from ai2_kit.core.log import get_logger
from ai2_kit.tool.deepmd import display_lcurve
from ..catalysis import AI2CAT_RES_DIR, ConfigBuilder, inspect_lammps_output

import matplotlib.pyplot as plt
from jupyter_formily import Formily
from typing import List, Tuple, Callable, Optional
import asyncio
import os
import glob


logger = get_logger(__name__)


class UiHelper:
    """
    Jupyter Widget for AI2CAT
    """

    def __init__(self) -> None:
        self.aimd_args = None
        self.train_args = None
        self.label_explore_args = None
        self.lammps_args = None

        self.system_file = None
        self.cp2k_basic_set_file = None
        self.cp2k_potential_file = None
        self.cp2k_parameter_file = None

        self.lammps_schema_path = os.path.join(AI2CAT_RES_DIR, 'gen-lammps.formily.json')
        self.selector_schema_path = os.path.join(AI2CAT_RES_DIR, 'selector.formily.json')

    def _set_default_system_file(self, args: dict):
        if self.system_file is not None:
            args['system_file'] = self.system_file

    def _update_default_system_file(self, args: dict):
        self.system_file = args.get('system_file', self.system_file)

    def _set_default_cp2k_basic_args(self, args: dict):
        if self.cp2k_basic_set_file is not None:
            args['basic_set_file'] = self.cp2k_basic_set_file
        if self.cp2k_potential_file is not None:
            args['potential_file'] = self.cp2k_potential_file
        if self.cp2k_parameter_file is not None:
            args['parameter_file'] = self.cp2k_parameter_file

    def _update_default_cp2k_basic_args(self, args: dict):
        self.cp2k_basic_set_file = args.get('basic_set_file', self.cp2k_basic_set_file)
        self.cp2k_potential_file = args.get('potential_file', self.cp2k_potential_file)
        self.cp2k_parameter_file = args.get('parameter_file', self.cp2k_parameter_file)


    def gen_aimd_config(self, out_dir: str, **default_value):
        if self.aimd_args is None:
            self.aimd_args = default_value
        self._set_default_system_file(self.aimd_args)
        self._set_default_cp2k_basic_args(self.aimd_args)

        schema = self._get_aimd_schema()
        options = _get_form_options('Provision AIMD Task')
        form = Formily(schema, options=options , default_value=self.aimd_args)

        form.display()
        async def _task():
            res = await wait_for_change(form, 'value')
            self.aimd_args = res['data']
            self._update_default_system_file(self.aimd_args)
            self._update_default_cp2k_basic_args(self.aimd_args)
            logger.info('form value: %s', self.aimd_args)
            try:
                logger.info('Start to generate AMID input files...')
                cp2k_kwargs: dict = self.aimd_args.copy()
                system_file = cp2k_kwargs.pop('system_file')
                config_builder = ConfigBuilder()
                config_builder.load_system(system_file)
                config_builder.gen_cp2k_input(out_dir=out_dir, **cp2k_kwargs, aimd=True)
                logger.info('Success!')  # TODO: Send a toast message
            except Exception as e:
                logger.exception('Failed!')  # TODO: Send a alert message
        asyncio.ensure_future(_task())


    def gen_train_vendors_config(self, out_dir: str, **default_value):
        if self.label_explore_args is None:
            self.label_explore_args = default_value
        self._set_default_system_file(self.label_explore_args)
        self._set_default_cp2k_basic_args(self.label_explore_args)

        schema = self._get_label_explore_schema()
        options = _get_form_options('Generate CP2K and Plumed Inputs')
        form = Formily(schema, options=options, default_value=self.label_explore_args)
        form.display()
        async def _task():
            res = await wait_for_change(form, 'value')
            self.label_explore_args = res['data']
            self._update_default_system_file(self.label_explore_args)
            self._update_default_cp2k_basic_args(self.label_explore_args)

            logger.info('form value: %s', self.label_explore_args)
            try:
                logger.info('Start to generate CP2K and plumed input files...')
                cp2k_kwargs: dict = self.label_explore_args.copy()
                system_file = cp2k_kwargs.pop('system_file')
                config_builder = ConfigBuilder()
                config_builder.load_system(system_file)
                config_builder.gen_cp2k_input(out_dir=out_dir, **cp2k_kwargs, aimd=False)
                config_builder.gen_plumed_input(out_dir=out_dir)
                logger.info('Success!')  # TODO: Send a toast message
            except Exception as e:
                logger.exception('Failed!')
        asyncio.ensure_future(_task())

    def gen_train_config(self, out_dir: str, **default_value):
        if self.train_args is None:
            self.train_args = default_value
        self._set_default_system_file(self.train_args)

        schema = self._get_train_schema()
        options = _get_form_options('Provision Training Workflow')
        form = Formily(schema, options=options, default_value=self.train_args)
        form.display()
        async def _task():
            res = await wait_for_change(form, 'value')
            self.train_args = res['data']
            self._update_default_system_file(self.train_args)
            logger.info('form value: %s', self.train_args)
            try:
                logger.info('Start to generate Training input files...')
                config_builder = ConfigBuilder()
                config_builder.load_system(self.train_args['system_file'])
                config_builder.gen_mlp_training_input(
                    out_dir=out_dir,
                    train_data=self.train_args.get('train_data', []),
                    explore_data=self.train_args.get('explore_data', []),
                    artifacts=self.train_args.get('artifacts', []),
                )
                config_builder.gen_deepmd_input(
                    out_dir=out_dir,
                    steps=self.train_args.get('steps'),  # TODO: provide default value
                )
                logger.info('Success!')  # TODO: Send a toast message
            except Exception as e:
                logger.exception('Failed!')  # TODO: Send a alert message
        asyncio.ensure_future(_task())

    def gen_lammps_config(self, out_dir: str, work_dir: str, **default_value):
        if self.lammps_args is None:
            self.lammps_args = default_value
        self._set_default_system_file(self.lammps_args)

        pattern = os.path.join(work_dir, '*/iters-*/train-deepmd/tasks/*/*.pb'  )
        dp_model_files = glob.glob(pattern)
        ensembles = ['nvt', 'nvt-i', 'nvt-a', 'nvt-iso', 'nvt-aniso', 'npt', 'npt-t', 'npt-tri', 'nve', 'csvr']

        schema = load_json(self.lammps_schema_path)
        # patch for FilePicker
        schema = merge_dict(schema, {'schema': {'properties': {
            'system_file': _get_file_picker(),
            'ensemble': {
                'enum': [{'children': [], 'label': e.upper(), 'value': e} for e in ensembles]
            },
            'dp_models': {
                'enum': [{'children': [], 'label': os.path.relpath(f, work_dir), 'value': os.path.abspath(f)}
                         for f in sorted(dp_model_files)]
            },
        }}}, quiet=True)
        options = _get_form_options('Provision LAMMPS Task')
        form = Formily(schema, options=options, default_value=self.lammps_args)
        form.display()

        async def _task():
            res = await wait_for_change(form, 'value')
            self.lammps_args = res['data']
            logger.info('form value: %s', self.lammps_args)
            try:
                kwargs = self.lammps_args.copy()
                system_file = kwargs.pop('system_file')
                logger.info('Start to generate LAMMPS input files...')
                config_builder = ConfigBuilder()
                config_builder.load_system(system_file)
                config_builder.gen_lammps_input(out_dir, **kwargs)
                config_builder.gen_plumed_input(out_dir=out_dir)
                logger.info('Success!')  # TODO: Send a toast message
            except Exception as e:
                logger.exception('Failed!')  # TODO: Send a alert message
        asyncio.ensure_future(_task())

    def inspect_deepmd_output(self, work_dir: str):
        pattern = os.path.join(work_dir, '*/iters-*/train-deepmd/tasks/*'  )
        dirs = glob.glob(pattern)
        options = [(os.path.relpath(d, work_dir), os.path.abspath(d)) for d in sorted(dirs)]
        fig_ax = plt.subplots(1, 2, figsize=(12, 4), constrained_layout=True)
        def _task(dir_path):
            display_lcurve(os.path.join(dir_path, 'lcurve.out'), fig_ax=fig_ax)
        self._gen_selector(title='Inspect DeepMD Result', label='Select DeepMD Task', options=options, cb=_task)

    def inspect_lammps_output(self, work_dir: str):
        pattern = os.path.join(work_dir, '*/iters-*/explore-lammps/tasks/*')
        dirs = glob.glob(pattern)
        options = [(os.path.relpath(d, work_dir), os.path.abspath(d)) for d in sorted(dirs)]
        fig_ax = plt.subplots(1, 2, figsize=(12, 4), constrained_layout=True)
        def _task(dir_path):
            inspect_lammps_output(dir_path, fig_ax=fig_ax)
        self._gen_selector(title='Inspect LAMMPS Result', label='Select LAMMPS Task', options=options, cb=_task)

    def _gen_selector(self, title: str, label: str, options: List[Tuple[str, str]], cb: Callable):
        schema = load_json(self.selector_schema_path)
        # patch options
        schema['schema']['properties']['selected']['enum'] = [{'children': [], 'label': opt[0], 'value': opt[1]} for opt in options]
        schema['schema']['properties']['selected']['title'] = label
        form = Formily(schema, options=_get_form_options(title))
        form.display()
        async def _task():
            res = await wait_for_change(form, 'value')
            logger.info('form data: %s', res)
            value = res['data']['selected']
            try:
                cb(value)
                logger.info('Success!')
            except Exception as e:
                logger.exception('Failed!')  # TODO: Send a alert message
        asyncio.ensure_future(_task())

    def _get_aimd_schema(self):
        schema_path = os.path.join(AI2CAT_RES_DIR, 'gen-cp2k-aimd.formily.json')
        schema = load_json(schema_path)
        # patch for FilePicker
        return merge_dict(schema, {'schema': {'properties': {
            'system_file':    _get_file_picker(),
            'basic_set_file': _get_file_picker(),
            'potential_file': _get_file_picker(),
            'parameter_file': _get_file_picker(),
        }}}, quiet=True)


    def _get_label_explore_schema(self):
        schema_path = os.path.join(AI2CAT_RES_DIR, 'gen-label-explore.formily.json')
        schema = load_json(schema_path)
        return merge_dict(schema, {'schema': {'properties': {
            'system_file':    _get_file_picker(),
            'basic_set_file': _get_file_picker(),
            'potential_file': _get_file_picker(),
            'parameter_file': _get_file_picker(),
        }}}, quiet=True)


    def _get_train_schema(self):
        schema_path = os.path.join(AI2CAT_RES_DIR, 'gen-training.formily.json')
        schema = load_json(schema_path)
        select_artifact_expr = "($deps[0] || []).map(item => ({label:item.key, value: item.key}))"
        schema = merge_dict(schema, {'schema': {'properties': {
            'system_file': _get_file_picker(),
            'train_data': _get_select_reactor(['artifacts'], select_artifact_expr),
            'explore_data': _get_select_reactor(['artifacts'], select_artifact_expr),
            'artifacts': {'items': { 'properties': {'artifact': {'properties': {
                'url': _get_file_picker(),
                'plumed_file': _get_file_picker(),
                'cp2k_file': _get_file_picker(),
            }}}}},
        }}}, quiet=True)
        return schema


def _get_select_reactor(deps: List[str], expr: str):
    return {"x-reactions": {
        "dependencies": deps,
        "fulfill": {
            "schema": {
                "enum": "{{%s}}" % expr,
            }
        }
    } }


def _get_file_picker(props: Optional[dict] = None) -> dict:
    if props is None:
        props = {}
    return {'x-component': 'FilePicker', 'x-component-props': props}


def _get_form_options(title: str):
    return {
        "modal_props": {
            "title": title,
            "width": "60vw",
            "style": {"max-width": "1200px"},
            "styles": {
                "body": {"max-height": "75vh", "overflow-y": "auto"}
            }
        }
    }


_UI_HELPER = None
def get_the_ui_helper():
    """
    Singleton for UiHelper
    """
    global _UI_HELPER  # pylint: disable=global-statement
    if _UI_HELPER is None:
        _UI_HELPER = UiHelper()
    return _UI_HELPER
